/*
 *       ____   ____      
 *    ___\   \ /   /____  
 *  _/ __ \   Y   // __ \ 
 *  \  ___/\     /\  ___/ 
 *   \___  >\___/  \___  >
 *       \/            \/ 
 *
 *  YaMoR simulator
 *
 *  Cyril Jaquier    
 *  Kevin Drapel
 *
 *  BIRG / LSL - EPFL
 *  http://birg.epfl.ch
 *
 * $Id$
 *
 */
 
#include <irrlicht.h>
#include <ode/ode.h>
#include <iostream>
#include <pthread.h>

#include "system.h"
#include "mmgr.h"  

#include "mesh.h"
#include "yamor.h"
#include "utils.h"
#include "robot.h"
#include "quakemap.h"
#include "xmlreader.h"
#include "button.h"
#include "xmlrpc.h"

using namespace irr;
using namespace core;
using namespace scene;
using namespace video;
using namespace io;
using namespace gui;
using namespace eve;

#pragma comment(lib, "Irrlicht.lib")
#define SYSTEM (System::getInstance())
	
Robot *myRobot; 
QuakeMap *myQuakeMap; 
XMLReader *myXMLReader;
OdeGlobalParameters odeGlobalParameters;


bool g_explodeRobot = false; 
bool g_printCameraPos = false; 
bool g_stopSimulation = false;
bool g_trackRobot = false; 

Button *btnRotation;
Button *btnTranslation;
Button *btnSimulation; 

int mouseX, mouseY; 

enum GUI_STATE {
	ROTATION = 0,
	TRANSLATION = 1,
	SIMULATION_PLAY = 2,
	IDLE = 100
};

GUI_STATE g_guiState = IDLE; 

vector3df g_robotTorque(0,0,0); 
vector3df g_robotTranslation(0,0,0);

void updatePhysics(Robot *robot);
void noticeGUI();  

/**
 * Irrlicht events receiver
 */
class MyEventReceiver : public IEventReceiver {
	public:
	
	virtual bool OnEvent(SEvent event) {
		
		if (event.EventType == irr::EET_MOUSE_INPUT_EVENT) {
			mouseX = event.MouseInput.X;
			mouseY = event.MouseInput.Y;
			
			if (event.MouseInput.Event == EMIE_LMOUSE_LEFT_UP) {
				SYSTEM->getCamera()->mouseUp(); 
				return true;
			} 

			if (event.MouseInput.Event == EMIE_LMOUSE_PRESSED_DOWN) {
				SYSTEM->getCamera()->mouseDown(event.MouseInput.X, event.MouseInput.Y); 
				noticeGUI(); 
				return true;
			} 
			
			if (event.MouseInput.Event == EMIE_MOUSE_MOVED && SYSTEM->getCamera()->isDragging()) {
				SYSTEM->getCamera()->mouseDown(event.MouseInput.X, event.MouseInput.Y); 
			}
		}
		
		if (event.EventType == irr::EET_KEY_INPUT_EVENT && !event.KeyInput.PressedDown) {
			switch (event.KeyInput.Key) {
				case KEY_SPACE:
					g_trackRobot = !g_trackRobot;
				return true; 
				case KEY_TAB:
					myRobot->tilt(odeGlobalParameters.guiSpeed*100.0); 
					updatePhysics(myRobot);
				return true; 
			
				case KEY_KEY_Q:
					SYSTEM->getCamera()->verticalStrafe(0.1);
				return true;
				
				case KEY_KEY_E:
					SYSTEM->getCamera()->verticalStrafe(-0.1); 
				return true;	
				
				case KEY_KEY_W:
					SYSTEM->getCamera()->move(0.1);
				return true;
				
				case KEY_KEY_S:
					SYSTEM->getCamera()->move(-0.1); 
				return true;	
				
				case KEY_KEY_A:
					SYSTEM->getCamera()->strafe(-0.1); 
				return true;	
				
				case KEY_KEY_D:
					SYSTEM->getCamera()->strafe(0.1); 
				return true;	
				
				case KEY_KEY_1:
					if (g_guiState == ROTATION) g_robotTorque.X += odeGlobalParameters.guiSpeed; 
					if (g_guiState == TRANSLATION) g_robotTranslation.X += odeGlobalParameters.guiSpeed*4; 
				return true; 
				
				case KEY_KEY_2:
					if (g_guiState == ROTATION) g_robotTorque.X -= odeGlobalParameters.guiSpeed; 
					if (g_guiState == TRANSLATION) g_robotTranslation.X -= odeGlobalParameters.guiSpeed*4;
				return true; 
				
				case KEY_KEY_3:
					if (g_guiState == ROTATION) g_robotTorque.Y += odeGlobalParameters.guiSpeed; 
					if (g_guiState == TRANSLATION) g_robotTranslation.Y += odeGlobalParameters.guiSpeed*4; 
				return true; 
				
				case KEY_KEY_4:
					if (g_guiState == ROTATION) g_robotTorque.Y -= odeGlobalParameters.guiSpeed; 
					if (g_guiState == TRANSLATION) g_robotTranslation.Y -= odeGlobalParameters.guiSpeed*4; 
				return true; 
				
				case KEY_KEY_5:
					if (g_guiState == TRANSLATION) g_robotTranslation.Z += odeGlobalParameters.guiSpeed; 
					if (g_guiState == ROTATION) g_robotTorque.Z += odeGlobalParameters.guiSpeed*4; 
				return true; 
				
				case KEY_KEY_6:
					if (g_guiState == TRANSLATION) g_robotTranslation.Z -= odeGlobalParameters.guiSpeed; 
					if (g_guiState == ROTATION) g_robotTorque.Z -= odeGlobalParameters.guiSpeed*4; 
				return true; 
					
				case KEY_KEY_7:
					if (g_guiState == TRANSLATION) g_robotTranslation = vector3df(0,0,0);  
					if (g_guiState == ROTATION) g_robotTorque = vector3df(0,0,0); 
				return true; 
				
				case KEY_KEY_Y:
				 	g_explodeRobot = true;
				return true; 
				case KEY_KEY_X:
					g_printCameraPos = !g_printCameraPos;
				return true; 
				case KEY_ESCAPE:
					g_stopSimulation = true;
				return true; 
			}			
		}	
		return false; 
	}
};

/**
 * Destroy the software and its allocated variables
 */
void destroy() {
	delete myQuakeMap; 
	delete myRobot; 
	SYSTEM->deleteInstance();
}

/**
 * Update the robot physics
 */
void updatePhysics(Robot *robot) {
	static double physicsInternalTimer = 0.0; 
	
	physicsInternalTimer += SYSTEM->getTimer()->DT;
	
	while(physicsInternalTimer > odeGlobalParameters.update) {
			SYSTEM->getODEWrapper()->step(odeGlobalParameters.step);
			physicsInternalTimer -= odeGlobalParameters.update;
	}

	robot->updatePhysics(); 	
	
	vector<Mesh*> meshList = SYSTEM->getMeshManager()->getMeshes(); 
	
	for (int n=0;n<meshList.size();n++) {
		meshList[n]->updateMeshAfterPhysics(); 
	}
}

/**
 * Setup the buttons in the GUI
 */
void setupGUI() {
	btnRotation = new Button(); 
	btnRotation->setup("data/button_rotation.tga", rect<s32>(5,5,5+64,5+64));
		
	btnTranslation = new Button(); 
	btnTranslation->setup("data/button_translation.tga", rect<s32>(5+64,5,5+64+64,5+64));
	
	btnSimulation = new Button(); 
	btnSimulation->setup("data/button_simulation.tga", rect<s32>(5+64+64,5,5+64+64+64,5+64));
}

/**
 * Draw the GUI
 */
void drawGUI() {
	btnRotation->draw(SYSTEM->getVideoDriver(), mouseX, mouseY,  (g_guiState==ROTATION) ? true : false );
	btnTranslation->draw(SYSTEM->getVideoDriver(), mouseX, mouseY,  (g_guiState==TRANSLATION) ? true : false );
	btnSimulation->draw(SYSTEM->getVideoDriver(), mouseX, mouseY, (g_guiState==SIMULATION_PLAY) ? true : false);
}

/**
 * Send information about mouse to the GUI
 */
void noticeGUI() {
	if (btnRotation->isMouseOver(mouseX, mouseY)) {
		g_guiState = ROTATION; 	
		g_robotTorque = vector3df(0,0,0); 
	}
	
	if (btnTranslation->isMouseOver(mouseX, mouseY)) {
		g_guiState = TRANSLATION;
		g_robotTranslation = vector3df(0,0,0);
	}
	
	if (btnSimulation->isMouseOver(mouseX, mouseY)) {
		g_guiState = SIMULATION_PLAY; 	
	} 
}

void disposeGUI() {
	delete btnRotation;
	delete btnSimulation;
	delete btnTranslation;
}
/**
 * Displace robot according to torques and forces applied on the main module
 */
void updateRobotLocation(Yamor *mainModule) {
	if (g_guiState == ROTATION) {
		dWorldSetGravity(SYSTEM->getODEWrapper()->getWorldID(), 0.0, 0.0, 0.0);
		dJointGroupEmpty(SYSTEM->getODEWrapper()->getFixedJointGroupID());
		dBodySetForce(mainModule->getMeshBlock()->getBodyID(), 0,0,0); 
		dBodyAddTorque(mainModule->getMeshBlock()->getBodyID(), g_robotTorque.X, g_robotTorque.Y, g_robotTorque.Z); 
		updatePhysics(myRobot); 
		SYSTEM->getODEWrapper()->addFixed(mainModule->getMeshBlock()->getBodyID(), 0); 
	}
	
	if (g_guiState == TRANSLATION) {
		dWorldSetGravity(SYSTEM->getODEWrapper()->getWorldID(), 0.0, 0.0, 0.0);
		dJointGroupEmpty(SYSTEM->getODEWrapper()->getFixedJointGroupID());
		dBodySetTorque(mainModule->getMeshBlock()->getBodyID(),0,0,0); 
		dBodyAddForce(mainModule->getMeshBlock()->getBodyID(), g_robotTranslation.X, g_robotTranslation.Y, g_robotTranslation.Z); 
		updatePhysics(myRobot); 
		SYSTEM->getODEWrapper()->addFixed(mainModule->getMeshBlock()->getBodyID(), 0); 
	}
	
	if (g_guiState == SIMULATION_PLAY) {
		dJointGroupEmpty(SYSTEM->getODEWrapper()->getFixedJointGroupID());  
		dWorldSetGravity(SYSTEM->getODEWrapper()->getWorldID(), 0.0,odeGlobalParameters.gravity, 0.0);
	}
}

/**
 * Main function
 */
int main(int argc, char* argv[])
{
	pthread_t serverThread;
	MyEventReceiver receiver;
	int exitValue;
	int irrWidth, irrHeight, irrFullscreen;
	
	/// read XML
	myXMLReader = new XMLReader(); 
	myXMLReader->readEveIrrlichtConfig("data/eve_config.xml", irrWidth, irrHeight, irrFullscreen);
	
	/// setup Irrlicht
	exitValue = SYSTEM->initIrrlicht(irrWidth, irrHeight, irrFullscreen, &receiver);
	
	if (exitValue == EXIT_FAILURE) {
		printf("FATAL ERROR - Couldn't initialize Irrlicht\n"); 
		exit(EXIT_FAILURE); 
	}	
	

	myXMLReader->readEveOdeConfig("data/eve_config.xml", odeGlobalParameters);
	
	/// setup ODE
	exitValue = SYSTEM->initODE(odeGlobalParameters);
	
	if (exitValue == EXIT_FAILURE) {
		printf("FATAL ERROR - Couldn't initialize ODE\n"); 
		exit(EXIT_FAILURE); 
	}
	
	/// setup XML-RPC server - default port value at 9999 (overidden by XML)
	vector3df robotPos, robotRot;
	double massModule, massLeveler; 
	
	int xmlRpcPort = 9999;
	
	std::string robotXMLName;
	
	myXMLReader->readEveConfigSetupWorld("data/eve_config.xml", &myQuakeMap, robotPos, robotRot, 
										                        xmlRpcPort, massModule, massLeveler, robotXMLName);

	myRobot = myXMLReader->readRobotDescription(robotXMLName, robotPos, robotRot, massModule, massLeveler); 
	
	/// start XML-RPC server
	XMLRpc::robot = myRobot; 
	XMLRpc::serverSetup(xmlRpcPort, "/tmp/eve_server.log"); 
	pthread_create( &serverThread, NULL, XMLRpc::serverThreadFunc, NULL);
		
	/// setup GUI
	setupGUI(); 
	
	/// start simulator
	SYSTEM->getTimer()->start(); 
	dWorldSetGravity(SYSTEM->getODEWrapper()->getWorldID(), 0.0, 0.0, 0.0);
	
	/// fix first module in space
	Yamor* firstModule = myRobot->getModule(1); 
	SYSTEM->getODEWrapper()->addFixed(firstModule->getMeshBlock()->getBodyID(), 0); 

	/// main loop
	while(SYSTEM->getDevice()->run() && !g_stopSimulation)
		if (SYSTEM->getDevice()->isWindowActive()) {
			SYSTEM->updateCamera();
			/// retrive important structures
			IVideoDriver *driver = SYSTEM->getVideoDriver();
			ISceneManager *smgr = SYSTEM->getSceneManager();
			SYSTEM->getTimer()->update(); 
		
			/// move the robot
			updateRobotLocation(firstModule);
			
		//	myRobot->getModule("424952470001")->setServoPosition(sin(SYSTEM->getTimer()->T*6.28)); 
		//	myRobot->getModule("424952470002")->setServoPosition(cos(SYSTEM->getTimer()->T*6.28)); 

			/// update physics and draw the scene
			updatePhysics(myRobot); 	

			driver->beginScene(true, true, 0);

			smgr->drawAll();
	
			/// special actions
			if (g_printCameraPos) {
				Utils::DebugVector3df(SYSTEM->getCameraIrrlicht()->getPosition(), "pos"); 
				Utils::DebugVector3df(SYSTEM->getCameraIrrlicht()->getTarget(), "trg"); 
			}
			
			if (g_explodeRobot) dJointGroupEmpty(SYSTEM->getODEWrapper()->getJointGroupID()); 

			/// draw the GUI on top of the scene
			drawGUI(); 
			
			/// end of scene
			driver->endScene();	

			SYSTEM->updateTitleBar();
			
			/// to prevent overheating of the CPU
			usleep(1); 
		}
	
		
	delete myXMLReader;
	
	/// dispose GUI	
	disposeGUI(); 
	
	/// dispose ODE
	SYSTEM->disposeODE();
	
	/// dispose Irrlicht and the system
	SYSTEM->disposeIrrlicht();
	destroy();

 ///	m_dumpMemoryReport("memory_report.txt", true);	
	return 0;
}

