/*
 *       ____   ____      
 *    ___\   \ /   /____  
 *  _/ __ \   Y   // __ \ 
 *  \  ___/\     /\  ___/ 
 *   \___  >\___/  \___  >
 *       \/            \/ 
 *
 *  YaMoR simulator
 *
 *  Cyril Jaquier    
 *  Kevin Drapel
 *
 *  BIRG / LSL - EPFL
 *  http://birg.epfl.ch
 *
 * $Id$
 *
 */
 
#include "system.h"
#include "mmgr.h"  

#include "mesh.h"
#include "utils.h"

namespace eve {
	
Mesh::Mesh() {
	m_bodyID = 0;
	m_mesh = NULL;
	m_mass = 1.0;
	m_centerOfMass = vector3df(0,0,0);
}

Mesh::~Mesh() {
}

/**
 * Load a mesh from a file
 */
void Mesh::loadMesh(std::string filename) {
	m_mesh = System::getInstance()->getSceneManager()->getMesh(filename.c_str());
}

/**
 * Add a new mesh
 */
void Mesh::addMesh() { 
	assert(m_mesh);
	m_node = System::getInstance()->getSceneManager()->addAnimatedMeshSceneNode(m_mesh);
	
	m_node->setMaterialFlag(EMF_FOG_ENABLE, false);
	m_node->setMaterialFlag(EMF_LIGHTING, true);

}

/**
 * Create an ODE BoxGeom from the bounding-box of the mesh
 */
void Mesh::createBoxGeomFromMesh(vector3df translate) {
	vector3df box = m_node->getBoundingBox().getExtent(); 
	vector3df scale = m_node->getScale();
	m_meshGeomID = dCreateBox(0, box.X*scale.X, box.Y*scale.Y, box.Z*scale.Z);
	dGeomSetPosition (m_meshGeomID, translate.X*scale.X, translate.Y*scale.Y, translate.Z*scale.Z);
	
}

/**
 * Create an ODE body from the mesh
 */
void Mesh::createBodyFromMesh(double mass, vector3df centerOfMass, bool isMovable) {

	m_mass = mass;
	m_centerOfMass = centerOfMass;

	// create the body
	if (!isMovable) {
		m_bodyID = 0; 
	} else {
		m_bodyID = dBodyCreate(System::getInstance()->getODEWrapper()->getWorldID());
	}
	
	// create transform space
	m_geomID = dCreateGeomTransform (System::getInstance()->getODEWrapper()->getSpaceID());
    
	// bind geom (useful ?) 
    dGeomTransformSetCleanup(m_geomID, 1);
	
	// insert collision geoms into transformation spaces
    dGeomTransformSetGeom(m_geomID, m_meshGeomID);
	
	vector3df pos = m_node->getPosition();
		
	// create the mass
	if (!isMovable) {
		    dGeomSetBody(m_geomID,0);
	} else {
			dGeomSetBody(m_geomID, m_bodyID); 	
			 
			// get the bounding box
			aabbox3d<f32> box = m_node->getBoundingBox();
			vector3df extent = box.getExtent();

			dMass bodyMass;
			vector3df scale = m_node->getScale();
			dMassSetBox(&bodyMass,1.0,(dReal)extent.X,(dReal)extent.Y*0.3,(dReal)extent.Z);
			dMassAdjust (&bodyMass, mass);	// mass of a Yamor module
			dMassTranslate(&bodyMass,centerOfMass.X,centerOfMass.Y,centerOfMass.Z); 
			
			dBodySetMass(m_bodyID, &bodyMass);
			dBodySetPosition(m_bodyID, pos.X, pos.Y, pos.Z); 
				 // set position	
			vector3df center = m_node->getBoundingBox().getCenter(); 
			//vector3df correction = center - pos; 
			dGeomSetPosition(m_geomID,pos.X, pos.Y, pos.Z);
				 
			// set rotation
			// TODO : check whether this should be applied to the '0' body
			dQuaternion q;
			vector3df eulerAngles = m_node->getRotation();
			Utils::EulerToQuaternion(eulerAngles, q);
			dGeomSetQuaternion(m_geomID, q); 
	  }
	 
}

/**
 * Update the mesh position/orientation after one ODE step
 */
void Mesh::updateMeshAfterPhysics() {
	
	vector3df pos;
	vector3df rot;

	if (m_bodyID!=0) {
		// get the new position of the ODE geometry
		dReal* ode_pos=(dReal*)dGeomGetPosition(m_geomID);
		// set the position at the scenenode
		pos.set((irr::f32)ode_pos[0],(irr::f32)ode_pos[1],(irr::f32)ode_pos[2]);

		m_node->setPosition(pos); 
		
		// get the rotation quaternion
		dQuaternion result;
		dGeomGetQuaternion(m_geomID, result);
		// convert it to eulerangles
		Utils::QuaternionToEuler(result,rot);
		// set the rotation 
		m_node->setRotation(rot);			
	}
}

} // namespace
