/*
 *       ____   ____      
 *    ___\   \ /   /____  
 *  _/ __ \   Y   // __ \ 
 *  \  ___/\     /\  ___/ 
 *   \___  >\___/  \___  >
 *       \/            \/ 
 *
 *  YaMoR simulator
 *
 *  Cyril Jaquier    
 *  Kevin Drapel
 *
 *  BIRG / LSL - EPFL
 *  http://birg.epfl.ch
 *
 * $Id$
 *
 */
  
#include "odewrapper.h"
#include "mmgr.h"  
#include "utils.h"

#define MAX_CONTACTS 200

namespace eve {
	
	
OdeWrapper *OdeWrapper::ptrToCurrent;
OdeGlobalParameters OdeWrapper::m_globalParameters; 
 dContact *OdeWrapper::m_contact; 
			
		 
OdeWrapper::OdeWrapper() {
}

OdeWrapper::~OdeWrapper() {
}

/**
 * Initialize the parameters for ODE
 */
void OdeWrapper::init(OdeGlobalParameters &odeGlobalParameters) {
	m_globalParameters = odeGlobalParameters; 
	
	// create world
	m_world = dWorldCreate();
	dWorldSetGravity(m_world, 0.0, m_globalParameters.gravity, 0.0);
	dWorldSetCFM(m_world,m_globalParameters.cfm); 
	dWorldSetERP(m_world,m_globalParameters.erp); 
	// create space
	m_space = dHashSpaceCreate(0); 

	
	// create contact group
	m_contactGroup = dJointGroupCreate(0); 
	m_jointGroup = dJointGroupCreate (0);
	m_fixedJointGroup = dJointGroupCreate (0);
		
	if (m_globalParameters.useQuickStep) {
		dWorldSetQuickStepNumIterations(m_world, m_globalParameters.quickStepIterations);
	}
	
	dWorldSetContactMaxCorrectingVel(m_world, m_globalParameters.contactMaxVelocity);
	dWorldSetContactSurfaceLayer(m_world, m_globalParameters.contactSurfaceLayer);
	
	m_contact = new dContact[m_globalParameters.maxContacts];
}

/**
 * Destroy ODE
 */
void OdeWrapper::dispose() {
	dJointGroupDestroy(m_contactGroup); 
	dJointGroupDestroy(m_jointGroup); 
	dJointGroupDestroy(m_fixedJointGroup); 
	dSpaceDestroy(m_space); 
	dWorldDestroy(m_world);
	delete[] m_contact;
}

/**
 * Compute one physics step
 */
void OdeWrapper::step(double stepSize) {
	// this is a trick to handle the callback with non-static members
	setCurrentWrapper(this); 
	dSpaceCollide(m_space, 0, &nearCollisionCallback);
	
	// step function
	if (m_globalParameters.useQuickStep) {
		dWorldQuickStep(m_world, stepSize); 
	} else {
		dWorldStep(m_world, stepSize); 
	}
	
	// remove all contact joints
	dJointGroupEmpty(m_contactGroup); 
}

/**
 * Add a fixed joint
 */
dJointID OdeWrapper::addFixed(dBodyID body1, dBodyID body2) {	
	dJointID id = dJointCreateFixed(m_world, m_fixedJointGroup); 
	
	dJointAttach(id, body1, body2);
	dJointSetFixed(id);

	return id;
}

/**
 * Collision callback that handles the collision joints
 */
void OdeWrapper::nearCollisionCallback(void* data, dGeomID o1, dGeomID o2) {
		dBodyID b1=dGeomGetBody(o1);
		dBodyID b2=dGeomGetBody(o2);
		
		//if (!(o1 == ptrToCurrent->getTerrainGeom() || o2 == ptrToCurrent->getTerrainGeom())) return;

			//	printf("%d %d %d\n" , o1, o2, ptrToCurrent->getTerrainGeom()); 
				
		if(b1 && b2 && dAreConnectedExcluding(b1,b2,dJointTypeContact)) return;
		if(b1 && b2 && dAreConnectedExcluding(b1,b2,dJointTypeHinge)) return;
		if(b1 && b2 && dAreConnectedExcluding(b1,b2,dJointTypeFixed)) return;

		for(int i=0;i<m_globalParameters.maxContacts;i++){
			m_contact[i].surface.mode = dContactBounce; 
			m_contact[i].surface.mu = m_globalParameters.surfaceFriction;//100.0;	// Coulomb (dInfinity = huge friction, 0 = frictionless)
			m_contact[i].surface.bounce=m_globalParameters.surfaceBounce;	// restitution : 1 = full restitution, 0 = full absorbtion (very soft material)
		}
		int numc = dCollide(o1,o2,MAX_CONTACTS,&m_contact[0].geom,sizeof(dContact));

		
		if(numc>0){
	
			for(int i=0;i<numc;i++){		
				dJointID c=dJointCreateContact(ptrToCurrent->getWorldID(),ptrToCurrent->getContactGroupID(),&m_contact[i]);
				dJointAttach(c,b1,b2);
			}
		}
}


} // namespace
