/*
 *       ____   ____      
 *    ___\   \ /   /____  
 *  _/ __ \   Y   // __ \ 
 *  \  ___/\     /\  ___/ 
 *   \___  >\___/  \___  >
 *       \/            \/ 
 *
 *  YaMoR simulator
 *
 *  Cyril Jaquier    
 *  Kevin Drapel
 *
 *  BIRG / LSL - EPFL
 *  http://birg.epfl.ch
 *
 * $Id$
 *
 */
  
 
#include "system.h"
#include "mmgr.h"  

#include "quakemap.h"	
#include "yamor.h"
#include "mesh.h"
#include "utils.h"

using namespace std; 

namespace eve {

QuakeMap::QuakeMap() {
	m_mesh = NULL;
	m_node = NULL;

}

QuakeMap::~QuakeMap() {
	if (m_vertices) delete[] m_vertices;
	if (m_indices) delete[] m_indices; 
}

/**
 * Load PK3 data file
 */
void QuakeMap::loadData(std::string pk3Filename) {
	System::getInstance()->getDevice()->getFileSystem()->addZipFileArchive(pk3Filename.c_str());
}

/**
 * Load a BSP level file from the PK3 or the virtual filesystem
 */
void QuakeMap::loadBSP(std::string bspFilename) {
	m_mesh =  System::getInstance()->getSceneManager()->getMesh(bspFilename.c_str());
	
	assert(m_mesh); 
	m_node =  System::getInstance()->getSceneManager()->addOctTreeSceneNode(m_mesh->getMesh(0), 0, -1, 1024);
	
	ITriangleSelector *s = System::getInstance()->getSceneManager()->createTriangleSelector(m_mesh->getMesh(0), m_node); 
	m_node->setTriangleSelector(s);
	s->drop();		
	
	m_vertices = new dVector3[s->getTriangleCount()*3];
    m_indices = new int[s->getTriangleCount()*3];
	
	//m_node->setVisible(false);
}

/**
 * Extract triangles from the Quake mesh
 */
vector<triangle3df> QuakeMap::extractTriangles() { 
	int countTrig = 0;
	
	ITriangleSelector *s = m_node->getTriangleSelector();
	triangle3df *triangles = new triangle3df[s->getTriangleCount()]; 
	
	s->getTriangles(triangles, s->getTriangleCount(), countTrig);
	
	vector<triangle3df> vTrig; 
	
	for (int n=0;n<countTrig;n++) {
		vTrig.push_back(triangles[n]); 
	}

	delete[] triangles; 
	
	return vTrig; 	
}

/**
 * Create the ODE geom from the Quake map
 */
dGeomID QuakeMap::createGeom(double quakeScale) {
	m_triangles = extractTriangles(); 

	int nbVertices = m_triangles.size()*3;
	
	int j=0;
	
	for (int n=0;n<m_triangles.size();n++) {
		
		m_triangles[n].pointA*=quakeScale; 
		m_triangles[n].pointB*=quakeScale;
		m_triangles[n].pointC*=quakeScale;
		
		m_vertices[j][0] = m_triangles[n].pointA.X; 
		m_vertices[j][1] = m_triangles[n].pointA.Y; 
		m_vertices[j][2] = m_triangles[n].pointA.Z; 
			
		m_vertices[j+1][0] = m_triangles[n].pointB.X; 
		m_vertices[j+1][1] = m_triangles[n].pointB.Y; 
		m_vertices[j+1][2] = m_triangles[n].pointB.Z; 	
		
		m_vertices[j+2][0] = m_triangles[n].pointC.X; 
		m_vertices[j+2][1] = m_triangles[n].pointC.Y; 
		m_vertices[j+2][2] = m_triangles[n].pointC.Z; 
		
		vector3df va = m_triangles[n].pointB - m_triangles[n].pointA; 
		vector3df vb = m_triangles[n].pointC - m_triangles[n].pointB; 
		
		vector3df badNormal = vb.crossProduct(va); 
		
		if (badNormal.dotProduct(m_triangles[n].getNormal().normalize()) < 0.0) {
			m_indices[j]=j;
			m_indices[j+1]=j+1;
			m_indices[j+2]=j+2;
		} else {
			m_indices[j]=j+2;
			m_indices[j+1]=j+1;
			m_indices[j+2]=j;
		}
		
		j+=3;
	}


	dTriMeshDataID data=dGeomTriMeshDataCreate();
    dGeomTriMeshDataBuildSimple(data,(dReal*)m_vertices, nbVertices, m_indices, nbVertices);	
    dGeomID geom = dCreateTriMesh(System::getInstance()->getODEWrapper()->getSpaceID(),data,0,0,0);
	dGeomSetData(geom,(void*)this);
	dGeomSetBody(geom,0);

	return geom; 
}


} // namespace
