/*
 *       ____   ____      
 *    ___\   \ /   /____  
 *  _/ __ \   Y   // __ \ 
 *  \  ___/\     /\  ___/ 
 *   \___  >\___/  \___  >
 *       \/            \/ 
 *
 *  YaMoR simulator
 *
 *  Cyril Jaquier    
 *  Kevin Drapel
 *
 *  BIRG / LSL - EPFL
 *  http://birg.epfl.ch
 *
 * $Id$
 *
 */
  
 
#include "system.h"
#include "mmgr.h"  

#include "yamor.h"
#include "robot.h"
#include "graph.h"
#include "graphnode.h"
#include "mesh.h"
#include "utils.h"
/*    
 */

namespace eve {

Robot::Robot(double massModule, double massLeveler) {
	m_isMovable = true;
	m_fixedModule = NULL; 
	m_moduleMass = massModule;
	m_levelerMass = massLeveler; 
}

Robot::~Robot() {
	for (int n=0;n<m_modules.size();n++) {
		delete m_modules[n];
	}
}

/**
 * Add a new YaMoR module
 */
void Robot::addModule(int address, std::string bluetoothAddr, vector3df position, vector3df rotation) {
	Yamor *yamor = new Yamor(address, bluetoothAddr);
	yamor->setYamorMass(m_moduleMass, m_levelerMass); 
	
	yamor->loadMeshes("data/module_centered.3ds", "data/leveler_centered.3ds"); 
	m_modules.push_back(yamor); 
	yamor->setPreODERotation(rotation); 
	yamor->setPreODEPosition(position); 
}

/**
 * Get module from its ID (in the scenegraph/robot tree)
 */
Yamor *Robot::getModule(int address) {
	for (int j=0;j<m_modules.size();j++) {
		if (m_modules[j]->getAddress()==address) return m_modules[j];
	}
	return NULL;
}

/**
 * Get module from its Bluetooth address
 */
Yamor *Robot::getModule(std::string bluetoothAddr) {
	for (int j=0;j<m_modules.size();j++) {
		if (m_modules[j]->getBluetoothAddress()==bluetoothAddr) return m_modules[j];
	}
	return NULL;
}

/**
 * Add a connection between two modules
 */
void Robot::addConnection(int address1, int address2, YAMOR_CONNECTOR co1, YAMOR_CONNECTOR co2, double angle) {
	Connection co;
	co.address1 = address1;
	co.address2 = address2;
	co.co1 = co1;
	co.co2 = co2;
	co.angle = angle;
	
	m_connections.push_back(co); 
}

/**
 * Align two modules to match a given connection
 */
void Robot::alignModules(int addressParent, int addressChild, YAMOR_CONNECTOR coParent, YAMOR_CONNECTOR coChild, double angle) {
	
	Yamor *moduleParent = getModule(addressParent); 
	Yamor *moduleChild = getModule(addressChild); 
	
	moduleChild->alignToConnector(-moduleParent->getConnectionNormal(coParent), moduleParent->getConnectionUp(coParent), 
							 	   moduleChild->getConnectionNormal(coChild),  moduleChild->getConnectionUp(coChild), angle);
				
	matrix4 mat = moduleParent->getMeshBlock()->getNode()->getRelativeTransformation(); 
	vector3df pt = moduleParent->getConnectionPosition(coParent); 
	
	mat = moduleChild->getMeshBlock()->getNode()->getRelativeTransformation(); 
	vector3df pt2 = moduleChild->getConnectionPosition(coChild);
	
	moduleChild->setPreODEPosition(moduleChild->getMeshBlock()->getNode()->getPosition()+(pt-pt2)); 

}

/**
 * Fix the modules when they are correctly aligned
 */
void Robot::fixModules(int addressParent, int addressChild, YAMOR_CONNECTOR coParent, YAMOR_CONNECTOR coChild) {
	Yamor *moduleParent = getModule(addressParent); 
	Yamor *moduleChild = getModule(addressChild); 
	moduleParent->setFixedConnection(moduleChild, coParent, coChild);
}

/**
 * Set the position of a module (before fixing them with ODE)
 */
void Robot::setPosition(int address, double position, bool invert) {
	Yamor *module = getModule(address);
	module->setServoPosition(position); 
}

/**
 * Add some random forces if the robot is stuck
 */
void Robot::tilt(float amount) {
	for (int n=0;n<m_modules.size();n++) {
		dBodyAddForce(m_modules[n]->getMeshBlock()->getBodyID(), amount*(frand()-0.5),amount*(frand()-0.5), amount*(frand()-0.5));
	}
}

/**
 * Generate the scene graph for the connections
 */
Graph *Robot::generateSceneGraph() {
	assert(m_modules.size()>0); 
	
	Graph *g = new Graph(); 
	
	for (int n=0;n<m_modules.size();n++) {
		g->createNode(m_modules[n]->getAddress()); 
	}
	
	GraphNode *g1;
	GraphNode *g2; 
	
	for (int n=0;n<m_connections.size();n++) {
		Connection co = m_connections[n];
		g->createConnection(g->getNodeFromAddress(co.address1),g->getNodeFromAddress(co.address2)); 
	}

	// we generate the tree, starting from the first node
	g->scan(g->getNode(0), g->getNode(0)); 
	
	return g; 
}

/**
 * Check if parent of a module is in the connection
 */
int Robot::isParentInConnection(int id1, int id2, vector<int> traversalOrder) {
	for (int n=0;n<traversalOrder.size();n++) {
		if (traversalOrder[n]==id1) return id1;
		if (traversalOrder[n]==id2) return id2;
	}
	
	// this must never happen
	assert(false); 
	return -1;
}

/**
 * Connect all modules together after the scenegraph is generated
 */
void Robot::connect() {
	Graph *g = generateSceneGraph(); 
	vector<int> order = g->m_traversalOrder; 

	for (int n=0;n<order.size();n++) {
		
		for (int j=0;j<m_connections.size();j++) {
			Connection co = m_connections[j];
			// we have a connection with this module
		
			if (co.address1 == order[n]) {
				int parentIs = isParentInConnection(order[n], co.address2, order);  
				// this link has already been added
				if (parentIs == co.address2) { printf("Not adding\n"); continue;} 

				 alignModules(co.address1,co.address2,co.co1,co.co2,co.angle);
			}
			
			// we have a connection with this module
			if (co.address2 == order[n]) {
				int parentIs = isParentInConnection(order[n], co.address1, order);
			
				// this link has already been added
				if (parentIs == co.address1) { printf("Not adding\n"); continue;}
				alignModules(co.address2,co.address1,co.co2,co.co1,co.angle);
			}
		
		}
	}
	
	generateODEData();
	
	for (int n=0;n<order.size();n++) {
		
		for (int j=0;j<m_connections.size();j++) {
			Connection co = m_connections[j];
			// we have a connection with this module
		
			if (co.address1 == order[n]) {
				int parentIs = isParentInConnection(order[n], co.address2, order);  
				// this link has already been added
				if (parentIs == co.address2) { printf("Not adding\n"); continue;} 

				 fixModules(co.address1,co.address2,co.co1,co.co2);
			}
			
			// we have a connection with this module
			if (co.address2 == order[n]) {
				int parentIs = isParentInConnection(order[n], co.address1, order);
			
				// this link has already been added
				if (parentIs == co.address1) { printf("Not adding\n"); continue;}
				fixModules(co.address2,co.address1,co.co2,co.co1);
			}
		
		}
	}
		
		
	delete g; 	
}

/**
 * Generate data for ODE
 */
void Robot::generateODEData() {
	for (int n=0;n<m_modules.size();n++) {
		m_modules[n]->generateODEData(true); 	
	}
}

/**
 * Returns the bounding box of the module
 */
aabbox3d<f32> Robot::getBoundingBox(double scaleRatio) {
	aabbox3d<f32> box; 
	
	for (int j=0;j<m_modules.size();j++) {
		box.addInternalBox(m_modules[j]->getMeshBlock()->getNode()->getBoundingBox()); 
		box.addInternalBox(m_modules[j]->getMeshLeveler()->getNode()->getBoundingBox()); 
	}
	 
	return box; 
}

/**
 * Update physics, change LEDs
 */
void Robot::updatePhysics() {
	static float randomCounter = 3.0; 
		
	for (int n=0;n<m_modules.size();n++) {

		m_modules[n]->updatePhysics(); 
		
		if (sin(randomCounter+n)>0.0) {
		m_modules[n]->setStateLED(false);
		} else {
			m_modules[n]->setStateLED(true);
		}
	}
	
	randomCounter -= 0.1;
	
	if (randomCounter<=0.0) {
		randomCounter = 1+rand()&15;
	}
}

/**
 * Set the motion status of the robot
 */
void Robot::isMovable(bool isMovable, int fixedModuleAddress) {
	m_fixedModule = getModule(fixedModuleAddress); 
	m_isMovable = isMovable; 
}


} // namespace
