/*
 *       ____   ____      
 *    ___\   \ /   /____  
 *  _/ __ \   Y   // __ \ 
 *  \  ___/\     /\  ___/ 
 *   \___  >\___/  \___  >
 *       \/            \/ 
 *
 *  YaMoR simulator
 *
 *  Cyril Jaquier    
 *  Kevin Drapel
 *
 *  BIRG / LSL - EPFL
 *  http://birg.epfl.ch
 *
 * $Id$
 *
 */
 
  
 
#include "system.h"
#include "mmgr.h"  

#include "utils.h"

namespace eve {
	
System* System::m_instance = NULL;

/**
 * Get singleton instance
 */
System *System::getInstance() {
	if (m_instance == NULL) {
		m_instance = new System(); 
	}
	
	return m_instance;
}

/**
 * Destroy singleton instance
 */
void System::deleteInstance() {
	if (m_instance) delete m_instance;  
}

System::System() {  
	m_odeWrapper = NULL; 
	m_irrlichtDevice = NULL; 
	m_irrlichtVideoDriver = NULL; 
	m_irrlichtSceneManager = NULL; 
	m_irrlichtCameraSceneNode = NULL; 
	
	m_timer = new Timer(); 
	m_camera = new Camera(); 
}

System::~System() {
	delete m_timer;
	delete m_camera; 
}

/**
 * Initialize Irrlicht, some parameters come from the XML or/and are overidden
 */
int System::initIrrlicht(int screenWidth, int screenHeight, bool fullscreen, IEventReceiver *receiver) {
	m_irrlichtDevice = createDevice(EDT_OPENGL, dimension2d<s32>(screenWidth, screenHeight), 16,  
	                                fullscreen, false, false, receiver);

	if (!m_irrlichtDevice) return EXIT_FAILURE;

	// init driver, scene manager and GUI
	m_irrlichtVideoDriver = m_irrlichtDevice->getVideoDriver();
	m_irrlichtSceneManager = m_irrlichtDevice->getSceneManager();

	m_irrlichtVideoDriver->setTextureCreationFlag(ETCF_ALWAYS_16_BIT, true);
	
	// add global camera
	m_irrlichtCameraSceneNode = m_irrlichtSceneManager->addCameraSceneNode(); 

	m_irrlichtCameraSceneNode->setPosition(vector3df( 19.921055,27.391312,79.015739));
	m_irrlichtCameraSceneNode->setTarget(vector3df(20.470171,26.660952,78.609482));
	m_irrlichtCameraSceneNode->setFOV(3.1415 / 1.15f);	
	m_irrlichtCameraSceneNode->setFarValue(120000.0f);
	m_irrlichtCameraSceneNode->setNearValue(0.1f);
	m_irrlichtCameraSceneNode->setInputReceiverEnabled(false); 
	
	m_meshManager = new MeshManager(); 
	
	return EXIT_SUCCESS;
}

/**
 * Dispose Irrlicht
 */
void System::disposeIrrlicht() {
	m_irrlichtDevice->drop();
	delete m_meshManager; 
}

/**
 * Initialize ODE
 */
int System::initODE(OdeGlobalParameters &odeGlobalParameters) {
	m_odeWrapper = new OdeWrapper(); 
	dRandSetSeed(1234); 
	m_odeWrapper->init(odeGlobalParameters); 
	
	return EXIT_SUCCESS;
}

/**
 * Dispose ODE
 */
void System::disposeODE() {
	m_odeWrapper->dispose(); 
	delete m_odeWrapper; 
	dCloseODE(); 
}

/**
 * Update title bar with useful information about performances
 */
void System::updateTitleBar() {
	static int lastFPS = -1;
	int fps = getVideoDriver()->getFPS();
	
	if (lastFPS != fps) {
		core::stringw str = L"EvE - YaMoR simulator [";
		str += getVideoDriver()->getName();
		str += "] fps -> ";
		str += fps;
		str += " primitives -> ";
		str += (int) getVideoDriver()->getPrimitiveCountDrawn();
		getDevice()->setWindowCaption(str.c_str());
		lastFPS = fps;
	}
}

/**
 * Update the camera
 */
void System::updateCamera() {
	getCameraIrrlicht()->setPosition(m_camera->m_position); 
	getCameraIrrlicht()->setTarget(m_camera->m_position + m_camera->m_direction); 
	getCameraIrrlicht()->updateAbsolutePosition(); 
}


} // namespace
