/*
 *       ____   ____      
 *    ___\   \ /   /____  
 *  _/ __ \   Y   // __ \ 
 *  \  ___/\     /\  ___/ 
 *   \___  >\___/  \___  >
 *       \/            \/ 
 *
 *  YaMoR simulator
 *
 *  Cyril Jaquier    
 *  Kevin Drapel
 *
 *  BIRG / LSL - EPFL
 *  http://birg.epfl.ch
 *
 * $Id$
 *
 */
  
 
#include "system.h"
#include <iostream>
#include "mmgr.h"  

#include "xmlreader.h"
#include "yamor.h"
#include "robot.h"
#include "utils.h"
#include "quakemap.h"


using namespace std; 

namespace eve {


XMLReader::XMLReader() {
}

XMLReader::~XMLReader() {
}

/**
 * Open a XML file
 */
TiXmlDocument *XMLReader::openFile(std::string filename) {
	TiXmlDocument *doc = new TiXmlDocument(filename.c_str());
	bool loadOkay = doc->LoadFile();
	
	if ( !loadOkay )
	{
		printf( "Could not load test file %s. Error='%s'. Exiting.\n", filename.c_str(), doc->ErrorDesc() );
		exit( 1 );
	}
	
	return doc; 
}

/**
 * Read robot description
 */
Robot *XMLReader::readRobotDescription(std::string filename, vector3df robotPos, vector3df robotRot, double massModule, double massLeveler) {
	TiXmlDocument *doc = openFile(filename); 
	
	TiXmlNode* nodeTemp = 0;
	
	TiXmlNode* nodeEve = doc->FirstChild("eve");
	assert(nodeEve);
	
	TiXmlNode* nodeModules = nodeEve->FirstChild("modules");
	assert(nodeModules);
	
	// create robot - TODO REPLACE EVERYTHING BY HASH
	Robot *robot = new Robot(massModule, massLeveler);
	
	vector<std::string> listAddr; 
	vector<std::string> listAlias; 
	int ID = 1; 
	
	map<std::string,int> aliasToID;
	 
	for(nodeTemp = nodeModules->FirstChild(); nodeTemp; nodeTemp = nodeTemp->NextSibling())
	{
		//address value="424952470001" alias="head"/>
		TiXmlElement *elemAddr = nodeTemp->ToElement();
		
		std::string addr(getStringAttribute(elemAddr, "value")); 
		std::string alias(getStringAttribute(elemAddr, "alias")); 
		cout << addr << alias << endl; 
		
		robot->addModule(ID, addr, robotPos, robotRot);
		listAddr.push_back(addr); 
		listAlias.push_back(alias); 
		
		aliasToID[alias] = ID;
		
		ID++;
	}	
	
	TiXmlNode* nodeConn= nodeEve->FirstChild("connections");
	assert(nodeConn);
	
	map<std::string,YAMOR_CONNECTOR> stringToConnector;
	
	stringToConnector["FACE_0"] = FACE_0;
	stringToConnector["FACE_1"] = FACE_1;
	stringToConnector["FACE_2"] = FACE_2;
	stringToConnector["FACE_3"] = FACE_3;
	stringToConnector["FACE_BOTTOM"] = FACE_BOTTOM;
	stringToConnector["FACE_LEVELER"] = FACE_LEVELER;
			
	for(nodeTemp = nodeConn->FirstChild(); nodeTemp; nodeTemp = nodeTemp->NextSibling())
	{
		//address value="424952470001" alias="head"/>
		TiXmlElement *elemCo = nodeTemp->ToElement();
		std::string moduleA(getStringAttribute(elemCo, "moduleA")); 
		std::string moduleB(getStringAttribute(elemCo, "moduleB")); 
		std::string faceA(getStringAttribute(elemCo, "faceA")); 
		std::string faceB(getStringAttribute(elemCo, "faceB")); 
		double angle(getDoubleAttribute(elemCo, "angle")); 
		
		robot->addConnection(aliasToID[moduleA], aliasToID[moduleB], stringToConnector[faceA], stringToConnector[faceB], angle); 
		cout << aliasToID[moduleA] << " " << aliasToID[moduleB] << faceA << faceB << angle << endl; 
	}	
	
	robot->connect(); 
	
	delete doc;
	
	return robot; 
}

/**
 * Read ODE config from XML
 */
void XMLReader::readEveOdeConfig(std::string filename, OdeGlobalParameters &odeParam) {
	TiXmlDocument *doc = openFile(filename); 
	
	TiXmlNode* nodeTemp = 0;
	vector3df initialPosition, initialRotation;
	
	TiXmlNode* nodeEve = doc->FirstChild("eve");
	assert(nodeEve);
	
	// read ODE config	
	TiXmlNode* nodeODE = nodeEve->FirstChild("ode_config");
	assert(nodeODE);
	
	odeParam.step = 1.0/(double) getIntAttribute(nodeODE->FirstChild("steps")->ToElement(), "value");
	odeParam.update = 1.0/(double) getIntAttribute(nodeODE->FirstChild("substeps")->ToElement(), "value");
	odeParam.guiSpeed = getDoubleAttribute(nodeODE->FirstChild("gui_move_rotation")->ToElement(), "value");
	odeParam.gravity = getDoubleAttribute(nodeODE->FirstChild("gravity")->ToElement(), "value");
	odeParam.cfm = getDoubleAttribute(nodeODE->FirstChild("cfm")->ToElement(), "value");
	odeParam.erp = getDoubleAttribute(nodeODE->FirstChild("erp")->ToElement(), "value");
	odeParam.useQuickStep = getIntAttribute(nodeODE->FirstChild("useQuickStep")->ToElement(), "value");
	odeParam.quickStepIterations = getIntAttribute(nodeODE->FirstChild("quickStepIterations")->ToElement(), "value");
	odeParam.contactSurfaceLayer = getDoubleAttribute(nodeODE->FirstChild("contactSurfaceLayer")->ToElement(), "value");
	odeParam.contactMaxVelocity = getDoubleAttribute(nodeODE->FirstChild("contactMaxVelocity")->ToElement(), "value");
	odeParam.maxContacts = getIntAttribute(nodeODE->FirstChild("maxContacts")->ToElement(), "value");

	odeParam.surfaceFriction = getDoubleAttribute(nodeODE->FirstChild("surface_friction")->ToElement(), "value");
	odeParam.surfaceBounce = getDoubleAttribute(nodeODE->FirstChild("surface_bounce")->ToElement(), "value");
		
	delete doc;
}

/**
 * Read Irrlicht config
 */
void XMLReader::readEveIrrlichtConfig(std::string filename, int &width, int &height, int &fullscreen) {
	TiXmlDocument *doc = openFile(filename); 
	
	TiXmlNode* nodeTemp = 0;
	
	TiXmlNode* nodeEve = doc->FirstChild("eve");
	assert(nodeEve);
	
	// read Irrlicht config	
	TiXmlNode* nodeODE = nodeEve->FirstChild("irrlicht_config");
	assert(nodeODE);
	
	width = getIntAttribute(nodeODE->FirstChild("width")->ToElement(), "value");
	height = getIntAttribute(nodeODE->FirstChild("height")->ToElement(), "value");
	fullscreen = getIntAttribute(nodeODE->FirstChild("fullscreen")->ToElement(), "value");
	
	delete doc;
}

/**
 * Read Eve config
 */
void XMLReader::readEveConfigSetupWorld(std::string filename, QuakeMap **ptrQuake, 
										vector3df &robotPos, vector3df &robotRot, int &serverPort, 
										double &massModule, double &massLeveler, std::string &robotXMLName) {
	TiXmlDocument *doc = openFile(filename); 
	
	TiXmlNode* nodeTemp = 0;
	vector3df initialPosition, initialRotation;
	
	TiXmlNode* nodeEve = doc->FirstChild("eve");
	assert(nodeEve);
	
	// read server description	
	TiXmlNode* nodeComm = nodeEve->FirstChild("communication");
	assert(nodeComm);
	serverPort = getIntAttribute(nodeComm->FirstChild("server")->ToElement(), "port");
	
	// read World description	
	TiXmlNode* nodeWorld = nodeEve->FirstChild("world");
	assert(nodeWorld); 
	
	//read camera node
	TiXmlNode* nodeCamera = nodeWorld->FirstChild("camera");
	assert(nodeCamera); 
	
	vector3df cameraPos = getVectorAttribute(nodeCamera->FirstChild("position")->ToElement()); 
	vector3df cameraTrg = getVectorAttribute(nodeCamera->FirstChild("target")->ToElement()); 
	
	System::getInstance()->getCamera()->m_position = cameraPos; 
	System::getInstance()->getCamera()->m_direction = cameraTrg - cameraPos; 
	
	// read Quake
	TiXmlNode* nodeQuake = nodeWorld->FirstChild("quakemap");
	assert(nodeQuake); 
	std::string pakFilename(getStringAttribute(nodeQuake->FirstChild("pak_file")->ToElement(), "name"));
	std::string bspFilename(getStringAttribute(nodeQuake->FirstChild("bsp_file")->ToElement(), "name"));
	double quakeScale = getDoubleAttribute(nodeQuake->FirstChild("scale")->ToElement(), "value");
	
	// setup quake
	*ptrQuake = new QuakeMap();
	(*ptrQuake)->loadData(pakFilename);
	(*ptrQuake)->loadBSP(bspFilename);
 
	(*ptrQuake)->getNode()->setScale(core::vector3df(quakeScale, quakeScale, quakeScale)); 
	dGeomID geomQuake = (*ptrQuake)->createGeom(quakeScale);
	System::getInstance()->getODEWrapper()->setTerrainGeom(geomQuake);
	
	// read robot configuration
	TiXmlNode* nodeRobot = nodeWorld->FirstChild("robot");
	assert(nodeRobot); 
	
	robotXMLName = getStringAttribute(nodeRobot->FirstChild("robot_file")->ToElement(), "name");
	robotPos.X  = getDoubleAttribute(nodeRobot->FirstChild("initial_position")->ToElement(), "x");
	robotPos.Y = getDoubleAttribute(nodeRobot->FirstChild("initial_position")->ToElement(), "y");
	robotPos.Z  = getDoubleAttribute(nodeRobot->FirstChild("initial_position")->ToElement(), "z");
	robotRot.X  = getDoubleAttribute(nodeRobot->FirstChild("initial_rotation")->ToElement(), "x");
	robotRot.Y = getDoubleAttribute(nodeRobot->FirstChild("initial_rotation")->ToElement(), "y");
	robotRot.Z  = getDoubleAttribute(nodeRobot->FirstChild("initial_rotation")->ToElement(), "z");

	massModule =  getDoubleAttribute(nodeRobot->FirstChild("module_mass")->ToElement(), "value");
	massLeveler =  getDoubleAttribute(nodeRobot->FirstChild("leveler_mass")->ToElement(), "value");
		
	// read objects & meshes config. 
	TiXmlNode* nodeObjs = nodeWorld->FirstChild("objects");
	assert(nodeObjs); 
	
	for(nodeTemp = nodeObjs->FirstChild(); nodeTemp; nodeTemp = nodeTemp->NextSibling())
	{
		// read mesh	
		std::string meshFilename(getStringAttribute(nodeTemp->FirstChild("mesh_file")->ToElement(), "name"));
		initialPosition.X  = getDoubleAttribute(nodeTemp->FirstChild("initial_position")->ToElement(), "x");
		initialPosition.Y = getDoubleAttribute(nodeTemp->FirstChild("initial_position")->ToElement(), "y");
		initialPosition.Z  = getDoubleAttribute(nodeTemp->FirstChild("initial_position")->ToElement(), "z");
		initialRotation.X  = getDoubleAttribute(nodeTemp->FirstChild("initial_rotation")->ToElement(), "x");
		initialRotation.Y = getDoubleAttribute(nodeTemp->FirstChild("initial_rotation")->ToElement(), "y");
		initialRotation.Z  = getDoubleAttribute(nodeTemp->FirstChild("initial_rotation")->ToElement(), "z");
		
		bool isMovable = getIntAttribute(nodeTemp->FirstChild("physics")->ToElement(), "movable"); 
		double mass = getDoubleAttribute(nodeTemp->FirstChild("physics")->ToElement(), "mass"); 
		double friction = getDoubleAttribute(nodeTemp->FirstChild("physics")->ToElement(), "friction"); 
		
		assert(mass>0.0);
		assert(friction>=0.0);

		// create the mesh
		Mesh *mesh = System::getInstance()->getMeshManager()->addMesh(meshFilename);
		
		mesh->getNode()->setPosition(initialPosition);
		mesh->getNode()->setRotation(initialRotation); 
		
		mesh->createBoxGeomFromMesh();
		mesh->createBodyFromMesh(mass, vector3df(1,0,2),isMovable);
	}	
	delete doc;
}

/**
 * Get vector attribute from XML
 */
vector3df XMLReader::getVectorAttribute(TiXmlElement *node) {
	vector3df v; 
	v.X = getDoubleAttribute(node, "x");
	v.Y = getDoubleAttribute(node, "y");
	v.Z = getDoubleAttribute(node, "z");
	return v; 
}

/**
 * Get string attribute from XML node
 */
const char *XMLReader::getStringAttribute(TiXmlElement *node, std::string name) {
	assert(node); 
	
	const char *result =  node->Attribute(name.c_str()); 
	
	if (result == NULL) {
		printf( "Could not read the 'string' attribute %s\n");  
		exit(1);
	}
	return result; 
}

/**
 * Get int attribute from XML node
 */
int XMLReader::getIntAttribute(TiXmlElement *node, std::string name) {
	assert(node); 
	
	int value; 
	int result = node->QueryIntAttribute(name.c_str(), (int*) &value); 
	
	if (result !=TIXML_SUCCESS) {
		printf( "Could not read the 'int' attribute %s (error %s)\n", name.c_str(), 
				(result==TIXML_WRONG_TYPE) ? "wrong type" : "no attribute found");
		exit(1);
	}
	
	return value; 
}

/**
 * Get double attribue from XML node
 */
double XMLReader::getDoubleAttribute(TiXmlElement *node, std::string name) {
	assert(node); 
	
	double value; 
	int result = node->QueryDoubleAttribute(name.c_str(), (double*) &value); 
	
	if (result !=TIXML_SUCCESS) {
		printf( "Could not read the 'double' attribute %s (error %s)\n", name.c_str(), 
				(result==TIXML_WRONG_TYPE) ? "wrong type" : "no attribute found");
		exit(1);
	}
	
	return value; 
}

} // namespace
