#include "system.h"
#include <iostream>
#include "mmgr.h"  

#include "xmlreader.h"
#include "yamor.h"
#include "robot.h"
#include "utils.h"
#include "quakemap.h"
#include "xmlrpc.h"

using namespace std; 

namespace eve {

Robot *XMLRpc::robot; 
xmlrpc_server_abyss_parms XMLRpc::serverparm;
xmlrpc_registry *XMLRpc::registryP;
xmlrpc_env XMLRpc::env;	
		
/**
 * Ping remote method. Adds two values together and returns the result via XML-RPC
 */
xmlrpc_value *XMLRpc::ping(xmlrpc_env *const env, xmlrpc_value *const param_array, void *const user_data) {
    xmlrpc_int32 x, y, z;

    xmlrpc_decompose_value(env, param_array, "(ii)", &x, &y);
    if (env->fault_occurred)
        return NULL;

    z = x + y;
	printf("PING? PONG!\n"); 
    return xmlrpc_build_value(env, "i", z);
}

/**
 * Set integer servo position via XML-RPC. 
 */
xmlrpc_value *XMLRpc::setServoPosition(xmlrpc_env *const env, xmlrpc_value *const param_array, void *const user_data) {
	const char *bluetoothAddr;
	xmlrpc_double position;
    xmlrpc_decompose_value(env, param_array, "(sd)", &bluetoothAddr, &position);
    if (env->fault_occurred) {
		printf("xmlrpc_value *setServoPosition - fault occured\n");
        return NULL;
	}
	
	robot->getModule(bluetoothAddr)->setServoPosition(position/90.0); 

    return xmlrpc_build_value(env, "i", 1);
}

/**
 * Set servo position via XML-RPC. 
 */
xmlrpc_value *XMLRpc::setServoIntPosition(xmlrpc_env *const env, xmlrpc_value *const param_array, void *const user_data) {
	const char *bluetoothAddr;
	xmlrpc_int position;
    xmlrpc_decompose_value(env, param_array, "(si)", &bluetoothAddr, &position);
    if (env->fault_occurred) {
		printf("xmlrpc_value *setServoIntPosition - fault occured\n");
        return NULL;
	}

	robot->getModule(bluetoothAddr)->setServoPosition(position/90.0); 

    return xmlrpc_build_value(env, "i", 1);
}


/**
 * Set all servos in one call
 */
xmlrpc_value * XMLRpc::setAllServos(xmlrpc_env *const env, xmlrpc_value *const param_array, void *const user_data) {
    unsigned int i, size = 0;
	xmlrpc_value *addr, *pos;
	
	const char *bluetoothAddr;
	xmlrpc_double position;
	
	size = xmlrpc_array_size(env, param_array);
	
	for (i=0; i<size; i+=2) {
	 addr = xmlrpc_array_get_item(env, param_array, i);
	 pos = xmlrpc_array_get_item(env, param_array, i+1);
	 xmlrpc_decompose_value(env, addr, "s", &bluetoothAddr);
	 if (env->fault_occurred) {
		 printf("xmlrpc_value bluetoothAddr fault occured\n");
		 return NULL;
	 }
	 
	 xmlrpc_decompose_value(env, pos, "d", &position);
	 
	 	 if (env->fault_occurred) {
		 printf("xmlrpc_value position fault occured\n");
		 return NULL;
	 }
	 
	 robot->getModule(bluetoothAddr)->setServoPosition(position/90.0); 
	}
     
    return xmlrpc_build_value(env, "i", 1);
}
 
/**
 * Setup XML-RPC server
 */
void XMLRpc::serverSetup(int port, std::string logfile) {
    xmlrpc_env_init(&env);

    registryP = xmlrpc_registry_new(&env);

	xmlrpc_registry_add_method(&env, registryP, NULL, "eve.ping", &ping, NULL);
	xmlrpc_registry_add_method(&env, registryP, NULL, "eve.setServoPosition", &setServoPosition, NULL);
	xmlrpc_registry_add_method(&env, registryP, NULL, "eve.setServoIntPosition", &setServoIntPosition, NULL); 	
	xmlrpc_registry_add_method(&env, registryP, NULL, "eve.setAllServos", &setAllServos, NULL); 
	
    serverparm.config_file_name = NULL;
    serverparm.registryP = registryP;
    serverparm.port_number = port;
    serverparm.log_file_name = logfile.c_str();
}
	
/**
 * Start the XML-RPC server thread
 */
void *XMLRpc::serverThreadFunc(void *ptr) {
	
    printf("Running XML-RPC server on port %d\n", serverparm.port_number);

    xmlrpc_server_abyss(&env, &serverparm, XMLRPC_APSIZE(log_file_name));
	
	while(true) {	
		usleep(1); 
	}
}


} // namespace
