/*
 *       ____   ____      
 *    ___\   \ /   /____  
 *  _/ __ \   Y   // __ \ 
 *  \  ___/\     /\  ___/ 
 *   \___  >\___/  \___  >
 *       \/            \/ 
 *
 *  YaMoR simulator
 *
 *  Cyril Jaquier    
 *  Kevin Drapel
 *
 *  BIRG / LSL - EPFL
 *  http://birg.epfl.ch
 *
 * $Id$
 *
 */
  
 
#include "system.h"
#include "mmgr.h"  

#include "yamor.h"
#include "mesh.h"
#include "utils.h"

/*          5
 *     .--------.                   0
 *    /          \           .____________.
 *   |    Servo   |          |            |
 *   |            |          |            |       ->
 *   |            |       1  |            3 ----> n = (1,0,0)
 *   |  Face "2"  |          |            |
 *   |            |          |____________|
 *   |____________|            [ Servo ]
 *          4                    Face "2"
 *                                 |
 *                                 |    ->
 *                                 v    n = (0,0,1)
 *                                                     
 *   Normal at faces (in local referential) : 
 *   Face 0 (servo) 			: (0,0,1)
 *   Face 1 					: (-1,0,0)
 *   Face 2 					: (0,0-1)
 *   Face 3 					: (1,0,0)
 *   Face 4 (Bottom of module) 	: (0,-1,0)
 *   Face 5 (Leveler) 			: (0,1,0)
 *
 *   The "up" of a module is the face 3. 
 *
 */

namespace eve {

Yamor::Yamor(int address, std::string bluetoothAddr) {
	// default values
	m_massBlock = 1.0;
	m_massLeveler = 1.0;
	m_centerOfMassBlock = vector3df(0.0,0.05,0.0); 
	m_centerOfMassLeveler = vector3df(0.0,0.0,0.0); 
	m_servoPosition = 0.0;
	m_servoMaxForce = 2000;
	m_servoMaxSpeed = 20.0; 
	
	m_address = address;
	m_bluetoothAddr = bluetoothAddr;
	
	// by default, no meshes
	m_meshBlock = NULL;
	m_meshLeveler = NULL;
	
	// this is to prevent displacements/rotations
	// after having set ODE primitives
	m_odeDataGenerated = false;
}

Yamor::~Yamor() {
	delete m_meshBlock;
	delete m_meshLeveler; 
}


void Yamor::setYamorMass(double moduleMass, double levelerMass) {
	m_massBlock = moduleMass;
	m_massLeveler = levelerMass; 
}

/**
 * Load the two meshes which are part of a module
 */
int Yamor::loadMeshes(std::string meshBlockFilename, std::string meshLevelerFilename) {
	assert(m_meshBlock == NULL); 
	assert(m_meshLeveler == NULL); 
	
	// allocate meshes
	m_meshBlock = new Mesh();
	m_meshLeveler = new Mesh(); 
	
	// load meshes
	m_meshBlock->loadMesh(meshBlockFilename);
	m_meshLeveler->loadMesh(meshLevelerFilename);
	
	// add to Irrlicht scenegraph
	m_meshBlock->addMesh(); 
	m_meshLeveler->addMesh(); 
	
	m_meshBlock->getNode()->setScale(vector3df(0.2, 0.2, 0.2)); 
	m_meshLeveler->getNode()->setScale(vector3df(0.2, 0.2, 0.2)); 
	
	// add Bluetooth LED :)
	m_ledNode = System::getInstance()->getSceneManager()->addBillboardSceneNode(m_meshBlock->getNode(), 
	dimension2d<f32>(0.01, 0.01), vector3df(0.18, 0.18, 0.05));
	m_ledNode->setMaterialFlag(EMF_LIGHTING, false);
	m_ledNode->setMaterialType(EMT_TRANSPARENT_ADD_COLOR);
	m_ledNode->setMaterialTexture(0, System::getInstance()->getVideoDriver()->getTexture("data/particle.bmp")); 
	
	return EXIT_SUCCESS;
}

/**
 * Set the state of the LED (fancy feature...)
 */
void Yamor::setStateLED(bool isVisible) {
	m_ledNode->setVisible(isVisible); 
}

/**
 * Generate ODE data
 */
int Yamor::generateODEData(bool isMovable) {
	m_meshBlock->createBoxGeomFromMesh(vector3df(0.0,-0.2,-0.0));
	m_meshLeveler->createBoxGeomFromMesh(vector3df(0.0,0.15,-0.0));;
		
	printf("%f %f ---------------", m_massBlock, m_massLeveler); 
	m_meshBlock->createBodyFromMesh(m_massBlock, m_centerOfMassBlock, isMovable);
	m_meshLeveler->createBodyFromMesh(m_massLeveler, m_centerOfMassLeveler, true);
		
	setServoHinge();

	m_odeDataGenerated = true;
	
	return EXIT_SUCCESS;
}

/**
 * Create and set a hinge joint
 */
void Yamor::setServoHinge() {
	OdeWrapper *ode = System::getInstance()->getODEWrapper(); 

	m_servoHingeJoint = dJointCreateHinge(ode->getWorldID(), ode->getJointGroupID()); 
	
	dJointAttach(m_servoHingeJoint, m_meshBlock->getBodyID(), m_meshLeveler->getBodyID());
	
	vector3df axis(0,0,1.0); 
	matrix4 mat = m_meshLeveler->getNode()->getRelativeTransformation(); 
	mat.rotateVect(axis); 
	
	dJointSetHingeAxis (m_servoHingeJoint, axis.X, axis.Y, axis.Z);

	vector3df anchor(0,0.0,0); 
	mat.transformVect(anchor); 
	

	dJointSetHingeAnchor (m_servoHingeJoint, anchor.X, anchor.Y, anchor.Z);
	
	dJointSetHingeParam(m_servoHingeJoint,dParamLoStop, -3.14*0.5);
	dJointSetHingeParam(m_servoHingeJoint,dParamHiStop, 3.14*0.5);
	
	dJointSetHingeParam(m_servoHingeJoint,dParamFMax,m_servoMaxForce);
}

/**
 * Create a fixed joint between two modules
 */
void Yamor::setFixedConnection(Yamor *other, YAMOR_CONNECTOR co1, YAMOR_CONNECTOR co2) {
	OdeWrapper *ode = System::getInstance()->getODEWrapper(); 
	assert(ode); 
	m_connectorFixedJoint = dJointCreateFixed(ode->getWorldID(), ode->getJointGroupID()); 

	if (co1 == FACE_LEVELER) {
		if (co2 == FACE_LEVELER) {
			dJointAttach(m_connectorFixedJoint, m_meshLeveler->getBodyID(), other->getMeshLeveler()->getBodyID());
		} else {
			dJointAttach(m_connectorFixedJoint, m_meshLeveler->getBodyID(), other->getMeshBlock()->getBodyID());
		}
	} else {
		if (co2 == FACE_LEVELER) {
			dJointAttach(m_connectorFixedJoint, m_meshBlock->getBodyID(), other->getMeshLeveler()->getBodyID());
		} else {
			dJointAttach(m_connectorFixedJoint, m_meshBlock->getBodyID(), other->getMeshBlock()->getBodyID());
		}
	}
		
	dJointSetFixed(m_connectorFixedJoint); 
}

/**
 * Set servo position (between -1 and 1)
 */
void Yamor::setServoPosition(double position) {
	if (position > 1.0) position = 1.0;
	if (position < -1.0) position = -1.0; 
	m_servoPosition = position;
}
			
/**
 * Update servo position with ODE
 */
void Yamor::updateServoPosition(bool invert) {

	// between -pi and pi
	double actualAngle = dJointGetHingeAngle(m_servoHingeJoint);
	
	// adding 1 degree of random noise
	double desiredAngle = m_servoPosition*3.1415*0.5;// + (frand()-0.5)*3.1415*0.5/180.0; 

	double newSpeed; 
	
	if (invert) {
		newSpeed = m_servoMaxSpeed*(desiredAngle - actualAngle);
	} else {
		newSpeed = -m_servoMaxSpeed*(desiredAngle - actualAngle);
	}

	dJointSetHingeParam(m_servoHingeJoint, dParamVel, newSpeed);
}

/**
 * Update physics
 */
void Yamor::updatePhysics() {
	updateServoPosition(true); 
	m_meshBlock->updateMeshAfterPhysics();
	m_meshLeveler->updateMeshAfterPhysics();
	
}

/**
 * Set position before fixing joints
 */
void Yamor::setPreODEPosition(vector3df pos) {
	assert(!m_odeDataGenerated);
	m_meshBlock->getNode()->setPosition(pos); 
	m_meshLeveler->getNode()->setPosition(pos); 
}

/**
 * Set rotation before fixing joints
 */
void Yamor::setPreODERotation(vector3df rot) {
	assert(!m_odeDataGenerated);
	m_meshBlock->getNode()->setRotation(rot); 
	m_meshLeveler->getNode()->setRotation(rot); 
}

/**
 * Align connectors
 */
void Yamor::alignToConnector(vector3df normalA, vector3df upA, vector3df normalB, vector3df upB, double angle) {
	
	normalA.normalize();
	upA.normalize();

	normalB.normalize();
	upB.normalize();

	// first : align the faces (opposite normals)
	vector3df v = normalA.crossProduct(normalB);
	if (v.getLength() < 0.00001) v = upA; 
	
	v.normalize(); 
	double theta = -acos(normalA.dotProduct(normalB)); 
	//theta = fmod(theta,3.14/2.0); 
	quaternion q;
	q.fromAngleAxis(theta, v); 
	
	matrix4 transformUp = q.getMatrix(); 
	transformUp.transformVect(normalB); 
    transformUp.transformVect(upA); 
	transformUp.transformVect(upB); 	
	
	//upA.normalize();
	//upB.normalize();
	//normalB.normalize(); 
	
	// second : rotate around the pivot 
	upA.normalize();
	upB.normalize(); 
	
	float dot = upA.dotProduct(upB);
	if (dot>1.0) dot=1.0; 
	if (dot<0.0) dot=0.0; 
	
	theta = acos(dot)*180/3.1415; 
	theta = fmod(theta,3.14/2.0); 
	//v = upA.crossProduct(upB);
	//if (v.getLength() < 0.00001) v = normalB; 
	
	theta -= angle; 
	
	printf("THETA : %f\n", theta); 
	quaternion q2;
	q2.fromAngleAxis(theta*3.1415/180.0, normalB); 
	
	q2*=q;

	vector3df euler;
	q2.toEuler(euler); 
	
	euler *= 180.0/3.1415;
	m_meshBlock->getNode()->setRotation(euler); 
	m_meshLeveler->getNode()->setRotation(euler); 
	
}

/**
 * Get the transformed/untransformed position of a connector
 */
vector3df Yamor::getConnectionPosition(YAMOR_CONNECTOR connectorNumber, bool transform) {
	vector3df position; 
	
	switch (connectorNumber) {
		case FACE_0:
			position = vector3df(0.0,-0.42,0.253);
		break;
		case FACE_1:
			position = vector3df(-0.260,-0.3,0.0);
		break;
		case FACE_2:
			position = vector3df(0,-0.42,-0.180);
	    break;
		case FACE_3:
			position = vector3df(0.261,-0.3,0);
	    break;
		case FACE_BOTTOM:
			position = vector3df(-0.0,-0.6,0.0);
	    break;
		case FACE_LEVELER: // TODO - pas juste, rotation !
			position = vector3df(0,0.397,0);
	    break;
    }
	
	if (transform) {
		matrix4 mat = m_meshBlock->getNode()->getRelativeTransformation();
		mat.transformVect(position); 	
	} else {
		return position;
	}
	
}

/**
 * Get the transformed/untransformed normal of a connector
 */
vector3df Yamor::getConnectionNormal(YAMOR_CONNECTOR connectorNumber, bool transform) {
	
	vector3df normal; 
	
	switch (connectorNumber) {
		case FACE_0:
			normal = vector3df(0,0,1);
		break;
		case FACE_1:
			normal = vector3df(-1,0,0);
		break;
		case FACE_2:
			normal = vector3df(0,0,-1);
	    break;
		case FACE_3:
			normal = vector3df(1,0,0);
	    break;
		case FACE_BOTTOM:
			normal = vector3df(0,-1,0);
	    break;
		case FACE_LEVELER: // TODO : PAS JUSTE ! -> rotation
			normal = vector3df(0,1,0);
	    break;
    }
	
	if (transform) {
		matrix4 mat = m_meshBlock->getNode()->getRelativeTransformation(); 
		mat.rotateVect(normal); 	
	} else {
		return normal;
	}
}

/**
 * Get the transformed/untransformed up vector of a connector
 */
vector3df Yamor::getConnectionUp(YAMOR_CONNECTOR connectorNumber, bool transform) {
	
	vector3df normal; 
	
	switch (connectorNumber) {
		case FACE_0:
			normal = vector3df(0,1,0);
		break;
		case FACE_1:
			normal = vector3df(0,1,0);
		break;
		case FACE_2:
			normal = vector3df(0,1,0);
	    break;
		case FACE_3:
			normal = vector3df(0,1,0);
	    break;
		case FACE_BOTTOM:
			normal = vector3df(1,0,0);
	    break;
		case FACE_LEVELER:
			normal = vector3df(-1,0,0);
	    break;
    }
	
	if (transform) {
		matrix4 mat = m_meshBlock->getNode()->getRelativeTransformation(); 
		mat.rotateVect(normal); 	
	} else {
		return normal;
	}
}


} // namespace
