/*
 *       ____   ____      
 *    ___\   \ /   /____  
 *  _/ __ \   Y   // __ \ 
 *  \  ___/\     /\  ___/ 
 *   \___  >\___/  \___  >
 *       \/            \/ 
 *
 *  YaMoR simulator
 *
 *  Cyril Jaquier    
 *  Kevin Drapel
 *
 *  BIRG / LSL - EPFL
 *  http://birg.epfl.ch
 *
 * $Id$
 *
 */
 
#ifndef _YAMOR_H_
#define _YAMOR_H_

#include <string>
#include <ode/ode.h>
#include <irrlicht.h>

/*          5
 *     .--------.                   2
 *    /          \           .____________.
 *   |    Servo   |          |            |
 *   |            |          |            |       ->
 *   |            |       1  |            3 ----> n = (1,0,0)
 *   |  Face "0"  |          |            |
 *   |            |          |____________|
 *   |____________|            [ Servo ]
 *          4                    Face "0"
 *                                 |
 *                                 |    ->
 *                                 v    n = (0,0,1)
 *                                                     
 *   Normal at faces (in local referential) : 
 *   Face 0 (servo) 			: (0,0,1)
 *   Face 1 					: (-1,0,0)
 *   Face 2 					: (0,0-1)
 *   Face 3 					: (1,0,0)
 *   Face 4 (Bottom of module) 	: (0,-1,0)
 *   Face 5 (Leveler) 			: (0,1,0)
 *
 */
 
using namespace irr;

namespace eve {

	enum YAMOR_CONNECTOR {
		FACE_0 = 0,
		FACE_1 = 1,
		FACE_2 = 2,
		FACE_3 = 3,
		FACE_BOTTOM = 4,
		FACE_LEVELER = 5
	};

	
	// external definition
	class Mesh;
	
	// class definition
	class Yamor {
		public:
			Yamor(int address, std::string bluetoothAddr);
			~Yamor(); 
			
			int loadMeshes(std::string meshBlockFilename, std::string meshLevelerFilename); 
			int generateODEData(bool isMovable);	
			void setServoPosition(double position);
			void updatePhysics();
			void setPreODEPosition(vector3df pos);
			void setPreODERotation(vector3df rot);
			void alignToConnector(vector3df normalA, vector3df upA, vector3df normalB, vector3df upB, double angle);
			
			Mesh *getMeshBlock() { return m_meshBlock; }
			Mesh *getMeshLeveler() { return m_meshLeveler; }
			
			int getAddress() { return m_address; }
			std::string getBluetoothAddress() { return m_bluetoothAddr;}
			
			vector3df getConnectionPosition(YAMOR_CONNECTOR connectorNumber, bool transform = true);
			vector3df getConnectionNormal(YAMOR_CONNECTOR connectorNumber, bool transform = true);
			vector3df getConnectionUp(YAMOR_CONNECTOR connectorNumber, bool transform = true);
		
			void setFixedConnection(Yamor *other, YAMOR_CONNECTOR co1, YAMOR_CONNECTOR co2);
			void setStateLED(bool isVisible);
			void setYamorMass(double moduleMass, double levelerMass);
		
		protected:
			void setServoHinge();
			void updateServoPosition(bool invert);
			
		protected:
			int	m_address;
			
			Mesh *m_meshBlock; 
			Mesh *m_meshLeveler;
			
			double m_massBlock;
			double m_massLeveler;
			vector3df m_centerOfMassBlock;
			vector3df m_centerOfMassLeveler;
			
			double m_servoPosition; 
			double m_servoMaxForce;
			double m_servoMaxSpeed;
			
			dJointID m_servoHingeJoint; 
			dJointID m_connectorFixedJoint;
			
			int m_odeDataGenerated;
			
			Yamor *m_parent; 
			
			IBillboardSceneNode *m_ledNode; 
			
			std::string m_bluetoothAddr; 
 	};	
	
	
} // namespace


#endif
